package com.github.huangp.learning.datastructure;

import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Deque;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import lombok.extern.slf4j.Slf4j;

/**
 * @author Patrick Huang
 *         <a href="mailto:pahuang@redhat.com">pahuang@redhat.com</a>
 */
@Slf4j
public class Graph<V> {
    private Multimap<V, V> adjacencyList = HashMultimap.create();
    private Map<V, Integer> entry = Maps.newHashMap();
    private Map<V, Integer> exit = Maps.newHashMap();
    private Map<V, Boolean> discovered = Maps.newHashMap();
    private Map<V, Boolean> processed = Maps.newHashMap();
    private Map<V, V> parent = Maps.newHashMap();

    private int numOfEdges = 0;
    private int time = 0;
    private final boolean directed;

    public Graph(boolean directed) {
        this.directed = directed;
    }


    public void insertEdge(V x, V y, boolean directed) {
        if (adjacencyList.containsKey(x)) {
            Collection<V> siblings = adjacencyList.get(x);
            siblings.add(y);
        } else {
            adjacencyList.put(x, y);
        }
        if (directed) {
            numOfEdges++;
        } else {
            insertEdge(y, x, true);
        }
    }

    public void print() {
        for (Map.Entry<V, Collection<V>> entry : adjacencyList
                .asMap().entrySet()) {
            V vertex = entry.getKey();
            Collection<V> siblings = entry.getValue();
            log.debug("node:{} degree:{}", vertex, siblings.size());
            for (V sibling : siblings) {
                log.debug("    edge:{}-{}", vertex, sibling);
            }
        }
    }

    public void BFS(V start) {
        Queue<V> queue = new ArrayDeque<V>();
        queue.add(start);
        discovered.put(start, true);
        parent.put(start, null);
        while (!queue.isEmpty()) {
            V next = queue.poll();
            preProcess(next);
            Collection<V> siblings = adjacencyList.get(next);
            for (V sibling : siblings) {
                if (unprocessed(sibling) || directed) {
                    processEdge(next, sibling);
                }
                if (undiscovered(sibling)) {
                    discovered.put(sibling, true);
                    parent.put(sibling, next);
                    queue.add(sibling);
                }
            }
            processed.put(next, true);
            postProcess(next);
        }
    }

    public List<V> findPath(V start, V end) {
        Deque<V> stack = new ArrayDeque<V>();
        stack.addFirst(end);
        V parentV = parent.get(end);
        while (parentV != null && parentV != start) {
            stack.addFirst(parentV);
            parentV = parent.get(parentV);
        }
        stack.addFirst(start);
        return ImmutableList.<V>builder().addAll(stack.iterator()).build();
    }

    protected void processEdge(V x, V y) {
        String edge = null;
        if (undiscovered(y) && !undiscovered(x)) {
            edge = "tree";
        } else if (!undiscovered(y) && !undiscovered(x)) {
            edge = "back";
        }
        log.debug("process edge {} -> {} [{}]", x, y, edge);
    }

    protected void preProcess(V vertex) {
        log.debug("processing {}", vertex);
    }

    protected void postProcess(V vertex) {
        log.debug("processed " + vertex);
    }

    public void DFS(V start) {
        preProcess(start);
        discovered.put(start, true);
        time++;
        entry.put(start, time);
        Collection<V> siblings = adjacencyList.get(start);
        for (V sibling : siblings) {
            if (undiscovered(sibling)) {
                parent.put(sibling, start);
                processEdge(start, sibling);
                DFS(sibling);
            } else if (unprocessed(sibling) || directed) {
                if (!parent.get(start).equals(sibling)) {
                    processEdge(start, sibling);
                }
            }
        }
        processed.put(start, true);
        postProcess(start);
        time++;
        exit.put(start, time);
    }

    private boolean unprocessed(V sibling) {
        return falseValue(processed.get(sibling));
    }

    private boolean undiscovered(V sibling) {
        return falseValue(discovered.get(sibling));
    }

    private static boolean falseValue(Boolean value) {
        return value == null || !value;
    }

    public void initSearch() {
        discovered.clear();
        processed.clear();
        entry.clear();
        exit.clear();
        parent.clear();
        time = 0;
    }

    public int getNumberOfVertices() {
        return adjacencyList.size();
    }

    public int getNumOfEdges() {
        return numOfEdges;
    }

    public Map<V, V> getParent() {
        return parent;
    }

    public Map<V, Integer> getEntry() {
        return entry;
    }

    public Map<V, Integer> getExit() {
        return exit;
    }
}
