package com.github.huangp.learning.datastructure;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import com.google.common.base.Objects;
import com.google.common.base.Preconditions;
import lombok.AllArgsConstructor;

/**
 * @author Patrick Huang
 *         <a href="mailto:pahuang@redhat.com">pahuang@redhat.com</a>
 */




public class HashMap<K, V> {
    private Entry[] backing = new Entry[20];
    private int size = 0;

    public V put(K key, V value) {
        Preconditions.checkNotNull(key);

        int index = indexOf(key.hashCode());

        Entry old = backing[index];
        if (old != null) {
            Entry cursor = old;
            while(cursor != null) {
                if (cursor.key.equals(key)) {
                    // need a replace
                    V oldValue = (V) cursor.value;
                    cursor.value = value;
                    return oldValue;
                }
                cursor = cursor.next;
            }
        }
        // pre-pend the entry to current bucket (old may be null)
        Entry<K, V> newEntry = new Entry<K, V>(key, value, old);
        backing[index] = newEntry;
        size++;

        return (V) old;
    }

    private int indexOf(int hash) {
        return hash % backing.length;
    }

    public boolean contains(K key) {
        int index = indexOf(key.hashCode());
        if (backing[index] == null) {
            return false;
        }
        Entry cursor = backing[index];
        while(cursor != null) {
            if (cursor.key.equals(key)) {
                return true;
            }
            cursor = cursor.next;
        }
        return false;
    }

    public boolean remove(K key) {
        int index = indexOf(key.hashCode());
        if (backing[index] == null) {
            return false;
        }
        Entry cursor = backing[index];
        Entry previous = null;
        while(cursor != null) {
            if (cursor.key.equals(key)) {
                // found it
                if (previous == null) {
                    backing[index] = cursor.next;
                } else {
                    previous.next = cursor.next;
                }
                size--;
                return true;
            }
            previous = cursor;
            cursor = cursor.next;
        }
        return false;
    }

    public int size() {
        return size;
    }

    public V get(K key) {
        int index = indexOf(key.hashCode());
        Entry cursor = backing[index];
        if (cursor == null) {
            return null;
        }
        while (cursor != null) {
            if (cursor.key.equals(key)) {
                return (V) cursor.value;
            }
            cursor = cursor.next;
        }
        return null;
    }

    public Set<K> keys() {
        Set<K> keys = new HashSet<K>(size);
        for (Entry<K, V> aBacking : backing) {
            Entry cursor = aBacking;
            while (cursor != null) {
                keys.add((K) cursor.key);
                cursor = cursor.next;
            }
        }
        return Collections.unmodifiableSet(keys);
    }

    @Override
    public String toString() {
        return Objects.toStringHelper(this)
                .add("backing", Arrays.toString(backing))
                .toString();
    }

    @AllArgsConstructor
    private static class Entry<K, V> {
        private K key;
        private V value;
        private Entry next;

        @Override
        public String toString() {
            return Objects.toStringHelper(this)
                    .add("key", key)
                    .add("value", value)
                    .add("next", next)
                    .toString();
        }
    }
}