package com.github.huangp.learning.datastructure;

import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Deque;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import com.google.common.collect.HashMultimap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Maps;
import com.google.common.collect.Multimap;
import lombok.extern.slf4j.Slf4j;

/**
 * @author Patrick Huang
 *         <a href="mailto:pahuang@redhat.com">pahuang@redhat.com</a>
 */
@Slf4j
public class WeightedGraph<V> {
    private Multimap<V, EdgeNode<V>> adjacencyList = HashMultimap.create();
    private Map<V, Integer> entry = Maps.newHashMap();
    private Map<V, Integer> exit = Maps.newHashMap();
    private Map<V, Boolean> discovered = Maps.newHashMap();
    private Map<V, Boolean> processed = Maps.newHashMap();
    private Map<V, V> parent = Maps.newHashMap();

    private final boolean directed;
    private int edgeNum = 0;

    public WeightedGraph(boolean directed) {
        this.directed = directed;
    }

    public void insertEdge(V x, V y, int weight) {
        insertEdge(x, y, weight, directed);
    }

    private void insertEdge(V x ,V y, int weight, boolean directed) {
        if (adjacencyList.containsKey(x)) {
            Collection<EdgeNode<V>> edgeNodes = adjacencyList.get(x);
            edgeNodes.add(new EdgeNode<V>(y, weight));
        } else {
            adjacencyList.put(x, new EdgeNode<V>(y, weight));
        }
        if (directed) {
            edgeNum++;
        } else {
            insertEdge(y, x, weight, true);
        }
    }

    public void initSearch() {
        entry.clear();
        exit.clear();
        discovered.clear();
        processed.clear();
        parent.clear();
    }

    public void BFS(V start) {
        Queue<V> queue = new ArrayDeque<V>();
        queue.add(start);
        while (!queue.isEmpty()) {
            V vertex = queue.poll();
            preProcess(vertex);
            discovered.put(vertex, true);
            Collection<EdgeNode<V>> edgeNodes = adjacencyList.get(vertex);
            for (EdgeNode<V> edgeNode : edgeNodes) {
                V sibling = edgeNode.vertex;
                if (!isProcessed(sibling) || directed) {
                    processEdge(vertex, edgeNode);
                }
                if (!isDiscovered(sibling)) {
                    discovered.put(sibling, true);
                    parent.put(sibling, vertex);
                    queue.add(sibling);
                }
            }
            postProcess(vertex);
            processed.put(vertex, true);
        }
    }

    private boolean isProcessed(V vertex) {
        Boolean value = processed.get(vertex);
        return value != null && value;
    }

    private void postProcess(V vertex) {
        log.debug("finished processing {}", vertex);
    }

    private void preProcess(V vertex) {
        log.debug("before processing {}", vertex);
    }

    private void processEdge(V x, EdgeNode<V> y) {
        log.debug("processing edge: {} -> {} ({})", x, y.vertex, y.weight);
    }

    private boolean isDiscovered(V vertex) {
        Boolean value = discovered.get(vertex);
        return value != null && value;
    }

    public int getNumOfEdges() {
        return edgeNum;
    }

    public Map<V, V> getParent() {
        return parent;
    }

    public List<V> findPath(V start, V end) {
        initSearch();
        BFS(start);
        Deque<V> stack = new ArrayDeque<V>();
        stack.addFirst(end);
        V parentV = parent.get(end);
        while (parentV != null && parentV != start) {
            stack.addFirst(parentV);
            parentV = parent.get(parentV);
        }
        stack.addFirst(start);
        return ImmutableList.<V>builder().addAll(stack.iterator()).build();
    }

    static class EdgeNode<V> {
        private final V vertex;
        private final int weight;

        EdgeNode(V vertex, int weight) {
            this.vertex = vertex;
            this.weight = weight;
        }

    }
}
